#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Classe de gestion d'une analyse d'un système de retraites.
"""
import pylab as pl
import os

class SimulateurAnalyse:
    def __init__(self, T, P, A, S, RNV, REV, Depenses, \
                 PIB, PensionBrut, \
                 scenarios, annees_EV, annees, annees_standard, \
                 scenarios_labels, scenarios_labels_courts, \
                 dir_image, ext_image):
        """
        Crée une analyse de simulateur de retraites.
        
        Paramètres :
            T: un dictionnaire, niveau des cotisations sociales
            P: un dictionnaire, niveau des pensions par rapport aux salaires
            A: un dictionnaire, âge moyen de départ à la retraite
            S: un dictionnaire, Situation financière du système de retraite en \% du PIB
            RNV: un dictionnaire, Niveau de vie des retraités par rapport à l'ensemble de la population
            REV: un dictionnaire, Durée de la vie passée à la retraite
            Depenses: un dictionnaire, Dépenses de retraites en % PIB
            PIB : le montant absolu du PIB (Milliard EUR)
            PensionBrut : la pension annuelle (brut) de droit direct moyenne (kEUR) 
            scenarios: une liste d'indices, les scénarios considérés
            annees_EV: une liste d'entiers, annees sur lesquelles on a l'espérance de vie
            annees: une liste d'entiers supérieurs ou égaux à 1, les scenarios consideres
            annees_standard: une liste d'entiers supérieurs ou égaux à 1, les années futures standard
            scenarios_labels : une liste de chaîne de caractères, les étiquettes 
                des scénarios économiques
            scenarios_labels_courts : une liste de chaîne de caractères, les étiquettes 
                courtes des scénarios économiques
            dir_image : le répertoire de sauvegarde des images
            ext_image : la liste des formats de sauvegarde des images
            
        Attributs :
            scenarios :
                Une liste d'entiers.
                la liste des scénarios considérés. 
                Ces scénarios sont des indices dans les tables de scénarios de 
                chomage, de croissance ainsi que les labels. 

            annees_EV :
                Une liste d'entiers.
                La liste des années de naissance pour lesquelles on a l'espérance de vie. 
    
            annees : 
                Une liste d'entiers.
                La liste des années sur lesquelles on fait les calculs. 
                Chaque année de cette liste est inférieure à l'année de l'horizon.
    
            annees_standard :
                Une liste d'entiers.
                La liste d'une sélection des années futures standard dans les calculs simplifiés. 
    
            T : 
                Un dictionnaire représentant une trajectoire. 
                Le taux de cotisations retraites.
            
            P : 
                Un dictionnaire représentant une trajectoire. 
                Le niveau moyen brut des pensions par rapport au 
                niveau moyen brut des salaires.
            
            A : 
                Un dictionnaire représentant une trajectoire. 
                L'âge effectif moyen de départ en retraite.
            
            S : 
                Un dictionnaire représentant une trajectoire. 
                Le solde financier du système de retraites en part de PIB. 

            RNV :
                Un dictionnaire représentant une trajectoire. 
                Niveau de vie des retraités par rapport à l'ensemble de la population. 

            REV :
                Un dictionnaire représentant une trajectoire. 
                Durée de la vie passée à la retraite.

            Depenses : 
                Un dictionnaire représentant une trajectoire. 
                Le montant des dépenses de retraites en part de PIB. 

            PIB : 
                Un dictionnaire représentant une trajectoire. 
                Le produit intérieur brut. 

            PensionBrut : 
                Un dictionnaire représentant une trajectoire. 
                Le montant annuel moyen brut de pension de droit direct. 
    
            scenarios_labels :
                La liste de chaîne de caractère décrivant les 
                scénarios pour chaque scénario 
                de la liste retournée par getScenarios().
    
            scenarios_labels_courts :
                La liste de chaîne de caractère courtes décrivant les 
                scénarios pour chaque scénario 
                de la liste retournée par getScenarios().

            labels_is_long :
                Un booléen. 
                True, si on utilise les labels longs dans les graphiques
        
            dir_image :
                Une chaîne de caractère. 
                Le répertoire de sauvegarde des images. 
                Par défaut, la chaîne de caractère vide. 

            ext_image :
                Une liste de chaînes de caractères. 
                Les types de fichier à générer par la méthode sauveFigure. 
                
            affiche_quand_ecrit :
                Un booléen.
                Si True, alors affiche un message quand la méthode sauveFigure 
                écrit un fichier.
        
            yaxis_lim :
                Un dictionnaire. 
                Les plages min et max pour l'axe des ordonnées 
                des variables en sortie de l'analyse. 
        

            liste_variables :
                Une liste de chaînes de caractère. 
                La liste des variables de l'analyse. 
                
            liste_legendes :
                Une liste de chaînes de caractère. 
                La liste des légendes des variables de l'analyse. 
        
        Exemple :
            simulateur = SimulateurRetraites()
            analyse = simulateur.pilotageCOR()
            analyse.dessineSimulation()
        """
        self.scenarios = scenarios
        self.annees_EV = annees_EV
        self.annees = annees
        
        # Liste des années dans le simulateur du COR
        self.annees_standard = annees_standard

        # initialisations diverses
        # chargement des donnees du COR pour les 6 scenarios
        
        self.T = T
        self.P = P
        self.A = A
        self.S = S
        self.RNV = RNV
        self.REV = REV
        self.Depenses = Depenses
        self.PIB = PIB
        self.PensionBrut = PensionBrut

        # Graphiques
        self.scenarios_labels = scenarios_labels
        self.scenarios_labels_courts = scenarios_labels_courts

        self.labels_is_long = True # True, si on utilise les labels longs
        
        self.dir_image = dir_image # répertoire pour les images

        self.ext_image = ext_image   # types de fichier à générer
        
        # Configure les plages min et max pour l'axe des ordonnées 
        # des variables standard en sortie du simulateur
        self.yaxis_lim = dict()
        self.yaxis_lim["S"] = [-2.0,2.0]
        self.yaxis_lim["RNV"] = [60.0,120.0]
        self.yaxis_lim["REV"] = [20.0,40.0]
        self.yaxis_lim["T"] = [25.0,40.0]
        self.yaxis_lim["A"] = [60,72]
        self.yaxis_lim["P"] = [25.0,55.0]
        self.yaxis_lim["Depenses"] = [11.0,15.0]

        self.affiche_quand_ecrit = True # Affiche un message quand on écrit un fichier
        
        self.liste_variables = ["S","RNV","REV","T","A","P","Depenses", "PIB", "PensionBrut"]
        self.liste_legendes=[u"Situation financière du système (% PIB)",
           u"Niveau de vie des retraités p/r à l'ensemble (%)",
           u"Proportion de la vie passée à la retraite (%)",
           u"Taux de cotisation de retraite (% PIB)",
           u"Age de départ effectif moyen à la retraite",
           u"Ratio (pension moyenne)/(salaire moyen) (%)",
           u"Dépenses de retraites (% PIB)", 
           u"Produit Intérieur Brut (Milliards EUR)", 
           u"Pension annuelle (brut) de droit direct (kEUR)"
        ]
        return None

    def setAfficheMessageEcriture(self, affiche_quand_ecrit):
        """
        Configure l'affichage d'un message quand on écrit un fichier
        
        Paramètres
            affiche_quand_ecrit : un booléen (par défaut = True)
        
        Exemple
            simulateur = SimulateurRetraites()
            analyse = simulateur.pilotageCOR()
            analyse.setAfficheMessageEcriture(False)
        """
        self.affiche_quand_ecrit = affiche_quand_ecrit
        return None
    
    def setImageFormats(self, ext_image):
        """
        Configure le format de sauvegarde des images
        
        Paramètres
            ext_image : une liste de chaînes de caractères (par defaut, ext_image=["png","pdf"])
        
        Exemple
            simulateur = SimulateurRetraites()
            analyse = simulateur.pilotageCOR()
            analyse.setImageFormats(["jpg"])
        """
        self.ext_image = ext_image
        return None

    def getImageFormats(self):
        """
        Retourne la liste des formats de sauvegarde des images.
        """
        return self.ext_image 

    
    def setLabelLongs(self, labels_is_long):
        """
        Configure la longueur des étiquettes
        
        Paramètres
            labels_is_long : un booléen, True si les labels longs sont utilisés (par défaut = True)
        
        Exemple
            simulateur = SimulateurRetraites()
            analyse = simulateur.pilotageCOR()
            analyse.setLabelLongs(False)
        """
        self.labels_is_long = labels_is_long
        return None

    def getLabelLongs(self):
        """
        Retourne le booléen de longueur des étiquettes
        
        Exemple :
            simulateur = SimulateurRetraites()
            analyse = simulateur.pilotageCOR()
            labels_is_long = analyse.getLabelLongs()
        """
        return self.labels_is_long 

    def setDirectoryImage(self, dir_image):
        """
        Configure le répertoire contenant les images
        
        Paramètres :
            dir_image : une chaîne de caractères, le répertoire contenant les images 
            (par défaut, dir_image="fig")
            exportées par sauveFigure.
        
        Exemple :
            simulateur = SimulateurRetraites()
            analyse = simulateur.pilotageCOR()
            analyse.setDirectoryImage("/tmp")
        """
        self.dir_image = dir_image
        return None

    def getDirectoryImage(self):
        """
        Retourne le répertoire contenant les images
        
        Exemple :
            simulateur = SimulateurRetraites()
            analyse = simulateur.pilotageCOR()
            dir_image = analyse.getDirectoryImage()
        """
        return self.dir_image

    def sauveFigure(self, f):
        """
        Sauvegarde l'image dans le répertoire
        
        Paramètres:
            f : une chaîne de caractères, le nom des fichiers à sauver
        
        Description :
            Sauvegarde l'image dans les formats définis. 
        
        Exemple:
            simulateur = SimulateurRetraites()
            analyse = simulateur.pilotageCOR()
            analyse.graphique("Depenses")
            analyse.sauveFigure("depenses")
        """
    
        for ext in self.ext_image:
            basefilename = f + "." + ext
            filename = os.path.join(self.dir_image,basefilename)
            if self.affiche_quand_ecrit:
                print("Ecriture du fichier %s" % (filename))
            pl.savefig(filename)
        return None
    
    def graphique(self, nom, v = None, taille_fonte_titre = 8, yaxis_lim = [], \
                  dessine_legende = False, scenarios_indices = None, 
                  dessine_annees = None):
        """
        Dessine un graphique associé à une variable donnée 
        pour tous les scénarios.
        
        Paramètres:
            nom : chaîne de caractère, nom de la variable
            v : variable à dessiner (par défaut, en fonction du nom)
            taille_fonte_titre : taille de la fonte du titre (par défaut, fs=8)
            yaxis_lim : une liste de taille 2, les bornes inférieures et supérieures 
            de l'axe des ordonnées
            dessine_legende : booleen, True si la légende doit être dessinée
            scenarios_indices : une liste d'entiers, la liste des indices des scénarios
            (par défaut, sc = range(1,7))
            dessine_annees : la liste des années à dessiner
        
        Description :
            Le nom peut être une égal à une des chaînes de caractères parmi 
            les chaînes suivantes : "T", "P", "A", "S", "RNV", "REV", 
            "Depenses", "PIB", "PensionBrut".
        
        Exemple:
            simulateur = SimulateurRetraites()
            analyse = simulateur.pilotageCOR()
            analyse.graphique("RNV")
            analyse.graphique("RNV", dessine_legende = True, scenarios_indices = range(1,5))
            analyse.graphique("RNV", dessine_annees = range(2020,2041))
            analyse.graphique("RNV", taille_fonte_titre = 14)
            analyse.graphique("B", simulateur.B)
        """
        
        if v is None:
            if nom=="T":
                v = self.T
            elif nom=="P":
                v = self.P
            elif nom=="A":
                v = self.A
            elif nom=="S":
                v = self.S
            elif nom=="RNV":
                v = self.RNV
            elif nom=="REV":
                v = self.REV
            elif nom=="Depenses":
                v = self.Depenses
            elif nom=="PIB":
                v = self.PIB
            elif nom=="PensionBrut":
                v = self.PensionBrut
            else:
                raise TypeError('Mauvaise valeur pour le nom : %s' % (nom))

        if scenarios_indices==None:
            scenarios_indices = self.scenarios
        
        if dessine_annees is not None:
            list_annees_dessin = dessine_annees
        else:
            if nom=="EV":
                list_annees_dessin=self.annees_EV
            else:
                list_annees_dessin=self.annees
    
        for s in scenarios_indices:
            if (nom=="S" or nom=="RNV" or nom=="T" or nom=="P" \
                or nom=="REV" or nom=="Depenses"):
                # Ce sont des % : multiplie par 100.0
                y = [ 100.0 * v[s][a] for a in list_annees_dessin ]
            else:
                y = [ v[s][a] for a in list_annees_dessin ]
            if (self.labels_is_long):
                label_variable = self.scenarios_labels[s]
            else:
                label_variable = self.scenarios_labels_courts[s]
            pl.plot(list_annees_dessin, y, label=label_variable )
    
        # titres des figures
        indice_variable = self.liste_variables.index(nom)
        titre_figure = self.liste_legendes[ indice_variable ]
           
        pl.title(titre_figure,fontsize=taille_fonte_titre)
        
        # Ajuste les limites de l'axe des ordonnées
        if yaxis_lim==[]:
            # If the use did not set the yaxis_lim
            if nom in self.yaxis_lim.keys():
                # If the variable name was found in the dictionnary
                yaxis_lim = self.yaxis_lim[nom]

        if yaxis_lim!=[]:
            pl.ylim(bottom=yaxis_lim[0],top=yaxis_lim[1])

        if dessine_legende:
            pl.legend(loc="best")
        return None
    
    def dessineSimulation(self, taille_fonte_titre=8):
        """
        Dessine les 9 graphiques "standards" 
        de la conjoncture pour tous les scénarios.
        
        Paramètres:
            taille_fonte_titre : taille de la fonte (par défaut, fs=8)
        
        Description :
            Dessine S, RNV, REV, T, A, P. 
        
        Exemple:
            simulateur = SimulateurRetraites()
            analyse = simulateur.pilotageCOR()
            analyse.dessineSimulation()
            analyse.dessineSimulation(taille_fonte_titre = 4)
        """
    
        for i in range(6):
            pl.subplot(3,2,i+1)
            nom= self.liste_variables[ i ]
            self.graphique(nom, taille_fonte_titre = taille_fonte_titre)
        pl.tight_layout(rect=[0, 0.03, 1, 0.95])
        return None
    
    def afficheVariable(self, v):
        """
        Affiche les valeurs d'une variable. 
        
        Paramètres :
            v : un dictionnaire, une trajectoire d'une des variables
        
        Exemple :
            simulateur = SimulateurRetraites()
            analyse = simulateur.pilotageCOR()
            analyse.afficheVariable(analyse.RNV)
        """
    
        for s in self.scenarios:
            print("Scenario",s,": ",self.scenarios_labels[s])
            for a in self.annees_standard:
                print("%d : %.3f"%(a, v[s][a]))
            print("")
        return None
            
    def afficheSolutionsSimulateurCOR(self):
        """
        Affiche les paramètres du simulateur. 
        
        Description :
            Les valeurs numériques affichées peuvent être utilisées 
            dans le simulateur du COR pour reproduire les simulations. 
        
        Exemple :
            simulateur = SimulateurRetraites()
            analyse = simulateur.pilotageCOR()
            analyse.afficheSolutionsSimulateurCOR()
        """
    
        print("Valeurs à rentrer sur le simulateur officiel du COR:")
        
        for s in self.scenarios:
            print("")
            print("Scenario",s,": ",self.scenarios_labels[s] )
            print("Annéee, Age,      Cotis., Pension:",)
            for a in self.annees_standard:
                print("%5d : %.1f ans, %.1f %%, %.1f %%"%(a, \
                                                         self.A[s][a], \
                                                         100*self.T[s][a], \
                                                         100*self.P[s][a]))
        return None
    
    def dessineLegende(self):
        """
        Crée un graphique présentant les légendes des graphiques.
        
        Exemple :
            simulateur = SimulateurRetraites()
            analyse = simulateur.pilotageCOR()
            analyse.dessineLegende()
        """
        # Juste les légendes
        pl.figure(figsize=(6,2))
        for s in self.scenarios:
            pl.plot(0., 0., label = self.scenarios_labels[s])
        pl.legend(loc="center")
        pl.ylim(bottom=0.0, top=0.7)
        pl.axis('off')
        return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Classe pour simuler l'étude d'impact de Janvier 2020.
"""
from scipy import interpolate
from copy import deepcopy

class EtudeImpact:
    def __init__(self, simulateur):
        """
        Créée un simulateur reproduisant l'étude d'impact.
        
        Paramètres :
            simulateur: un SimulateurRetraites.
        
        Exemple :
            simulateur = SimulateurRetraites()
            etudeImpact = EtudeImpact(simulateur)
            etudeImpact.calcule()
            Ds = etudeImpact.getDepenses()
            Ss = etudeImpact.getSolde()
            As = etudeImpact.getAge()
            
        Attributs :
            simulateur : 
                un SimulateurRetraites
            depenses_annee_transition :
                un entier. 
                L'année de transition de l'étude d'impact.
            depenses_annee_extrapolation :
                un entier.
                La première année d'extrapolation au delà des données 
                de l'étude d'imact. 
            depenses_annees : 
                une liste d'entiers. 
                La liste des années où le montant des dépenses est connu 
                dans l'étude d'impact. 
            depenses_valeurs : 
                une liste de flottants.
                La liste des montants des dépenses d'après le tableau 39 
                de l'étude d'impact. 
            analyse : 
                un SimulateurAnalyse. 
                L'analyse du pilotage du COR. 
            Ds :
                un dictionnaire. 
                Le montant des dépenses dans chaque scénario, 
                pour chaque année. 
            Ss :
                un dictionnaire. 
                Le solde financier dans chaque scénario, 
                pour chaque année.
            As :
                un dictionnaire. 
                L'âge effectif moyen de départ en retraite dans chaque scénario, 
                pour chaque année. 
            premiere_generation : 
                Un entier. 
                L'année de naissance de la première génération affectée 
                par la transition. 
            age_premiere_generation :
                Un flottant. 
                L'âge moyen de départ à la retraite de la première 
                génération affectée par la transition. 
            derniere_generation :
                Un entier. 
                L'année de naissance la plus grande détaillée dans l'étude 
                d'impact relatif à l'âge effectif moyen de départ en retraite. 
            age_derniere_generation :
                Un flottant. 
                L'âge effectif moyen de départ en retraite de la dernière 
                génération évoquée dans l'étude d'impact.
            age_annee_transition : 
                un entier. 
                L'année où la première génération de la transition 
                part en retraite. 
            age_annee_extrapolation : 
                un entier. 
                L'année où la dernière génération évoquée dans l'étude d'impact 
                part en retraite. 
            solde_annee_transition : 
                Un entier. 
                La première année de transition pour le solde.
            solde_annee_equilibre : 
                Un entier. 
                La première année d'équilibre pour le solde.
            epaisseur_ligne : 
                Un entier. 
                L'épaisseur de la ligne pour reproduire graphiquement 
                l'étude d'impact.
            couleur_HR : 
                Une chaîne de caractère. 
                La couleur d'une courbe "Hors réforme" ou "Avant réforme".
            couleur_SUR : 
                Une chaîne de caractère. 
                La couleur d'une courbe "SUR" ou "Après réforme".
            
        """
        # Liste des années dans le simulateur du COR
        self.simulateur = simulateur
        
        # Paramètres pour le calcul des dépenses
        self.depenses_annee_transition = 2020
        self.depenses_annee_extrapolation = 2050
        self.depenses_annees =  [2025,  2030,  2040,  2050,  2060,  2070]
        self.depenses_valeurs = [0.136, 0.135, 0.133, 0.129, 0.1275, 0.126]
        
        self.analyse = self.simulateur.pilotageCOR()
        self.Ds = deepcopy(self.analyse.Depenses)
        self.Ss = deepcopy(self.analyse.S)
        self.As = deepcopy(self.analyse.A)
        
        # Paramètres pour le calcul des âges
        # Paramètres de l'interpolation linéaire
        self.premiere_generation = 1975
        self.age_premiere_generation = 63.6
        self.derniere_generation = 2000 # Graphique 49 : 1990
        self.age_derniere_generation = 65.2 # Graphique 49 : 64.83
        self.age_annee_transition = int(self.premiere_generation + self.age_premiere_generation) # 2038
        self.age_annee_extrapolation = int(self.derniere_generation + self.age_derniere_generation) # 2065
        
        # Paramètre pour le calcul du solde
        self.solde_annee_transition = 2020
        self.solde_annee_equilibre = 2027 # Première année d'équilibre strict
        
        # Préférences graphiques de l'étude d'impact
        self.epaisseur_ligne= 3
        self.couleur_HR = "cornflowerblue"
        self.couleur_SUR = "orange"

        return None

    def calcule(self):
        """
        Calcule la trajectoire dans l'étude d'impact.
        Le pilotage est réalisé à partir de l'âge moyen de départ en retraite, 
        des dépenses et du solde financier. 
        """
        self.calculeDepenses()
        self.calculeAgeDepartRetraite()
        self.calculeSolde()
        
        analyse = self.simulateur.pilotageParSoldeAgeDepenses(Scible=self.Ss, Acible=self.As, Dcible=self.Ds)
        return analyse
    
    def calculeDepenses(self):
        """
        Calcule la trajectoire des dépenses dans l'étude d'impact.
        Source : Tableau 39 de l'étude d'impact, page 176. 
        Méthode interpolation quadratique dans les 
        données de la table. 
        """        

        # Crée une liste d'années et une liste de dépenses
        # Crée une liste initiale : la première valeur sera remplacée pour chaque 
        # scénario par la suite.
        annees =   [self.depenses_annee_transition]
        for a in self.depenses_annees:
            annees.append(a)
        valeur_factice = -1.0
        depenses = [valeur_factice]
        for d in self.depenses_valeurs:
            depenses.append(d)            
        # Boucle sur les scénarios
        for s in self.simulateur.scenarios:
            # Affecte la dernière dépense constatée dans le scénario
            depenses[0] = self.analyse.Depenses[s][self.depenses_annee_transition]
            
            # Crée l'interpolateur
            depenses_interpolateur = interpolate.interp1d(annees, depenses, kind="quadratic")
            
            # Corrige les dépenses
            for a in self.simulateur.annees_futures:
                if (a>=self.depenses_annee_transition):
                    self.Ds[s][a] = depenses_interpolateur(a)

        return None

    def ageDepartParAnnee(self, a):
        """
        Calcule l'âge de départ pour une année de départ en retraite donnée.

        Paramètres
            a : année de départ en retraite
        
        Description :
            Retourne l'âge de départ en retraite.
        """
        # Interpolation linéaire inverse
        Age1 = self.age_premiere_generation
        Age2 = self.age_derniere_generation
        an1 = self.premiere_generation
        an2 = self.derniere_generation
        As = (Age2 * (a - an1) + Age1 * (an2 - a)) / ((an2 - an1) + (Age2 - Age1))
        return As

    def ageDepartParGeneration(self, an):
        """
        Calcule l'âge de départ pour une année de naissance donnée. 
        
        Paramètres :
            an : année de naissance
        
        Description :
            Retourne l'âge de départ en retraite
        """
        # Interpolation linéaire inverse
        Age1 = self.age_premiere_generation
        Age2 = self.age_derniere_generation
        an1 = self.premiere_generation
        an2 = self.derniere_generation
        As = Age1 * (an2 - an) / (an2 - an1) + Age2 * (an - an1) / (an2 - an1)
        return As

    def calculeAgeDepartRetraite(self):
        """
        Calcule l'âge de départ en retraite de l'étude d'impact (Janvier 2020).
        Source : graphique 73 page 139 de l'étude d'impact de Janvier 2020. 
        Méthode : interpolation linéaire, puis inversion formelle des équations 
        pour obtenir l'âge de départ en retraite en fonction de l'année 
        de départ en retraite. 
        
        Note : dans la première version, nous utilisions le 
        graphique 49 page 139 de l'étude d'impact de Janvier 2020. 
        Mais le graphique 49 s'arrête en 1990 alors que le graphique 73 
        présente jusqu'à 2000.
        """
    
        for s in self.simulateur.scenarios:
            for a in self.simulateur.annees_futures:
                if (a>=self.age_annee_transition):
                    self.As[s][a] = self.ageDepartParAnnee(a)
                    
        return None
    
    def calculeSolde(self):
        """
        Calcul le solde de l'étude d'impact (Janvier 2020).
        Source : graphique 63, page 180 "Solde du système de retraites, avant et après réforme"
        Données : courbe pointillé orange jusqu'à 2027, puis 
        courbe orange en trait plein.
        Method : interpolation linéaire
        """
        annees = [self.solde_annee_transition, 2027, 2070]
        
        # Met à jour le solde
        for s in self.simulateur.scenarios:
            dernier_solde_constate = self.analyse.S[s][self.solde_annee_transition]
            soldes = [dernier_solde_constate, 0.0, 0.0]
    
            # Crée l'interpolateur de solde financier   
            soldes_interpolateur = interpolate.interp1d(annees, soldes)
            for a in self.simulateur.annees_futures:
                if (a>=self.solde_annee_transition):
                    self.Ss[s][a] = soldes_interpolateur(a)
        
        return None
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import openturns as ot
import scipy as sp

class FonctionPension(ot.OpenTURNSPythonFunction):
    def __init__(self, simulateur, annee, verbose = False):
        """
        Crée un modèle de pension.
        
        Paramètres :
            simulateur : un SimulateurRetraite
            annee : l'année de calcul de P
            verbose : si vrai, affiche des variables intermédiaires durant l'évaluation
        
        Description :
            Crée un modèle de pension permettant d'évaluer le 
            ratio (pension moyenne) / (salaire moyen) pour une année donnée.
        
            Les entrées de la fonction sont "S", "D", "As", "F", "TauC" 
            et la sortie est "P". 
            
            * S : le solde financier du système de retraites (% PIB)
            * D : le montant des dépenses (% PIB)
            * As : l'âge moyen de départ à la retraite défini par l'utilisateur
            * F  : facteur d'élasticité de report de l'âge de départ 
                (par exemple F=0.5)
            * TauC : le taux de chômage (par exemple TauC = 4.5)
            * P : le rapport entre le montant moyen des pensions et le 
                 montant moyen des salaires
        
            Les autres paramètres  sont ceux du simulateur, 
            dans le scénario central (i.e. s = 1) à l'année considérée. 
        
            Le modèle calcule NC, dP, B, G, A, NR en fonction du taux 
            de chômage TauC par interpolation dans les données 
            du COR. 
            
        Attributs :
            simulateur : 
                un SimulateurRetraites
            annee : 
                un entier. L'année de calcul.
            verbose : 
                un booléen. Si vrai, affiche les calculs intermédiaires.
            interpolateur_NC : 
                une fonction. 
                L'interpolateur du nombre de cotisants. 
            interpolateur_dP : 
                une fonction. 
                L'interpolateur des autres dépenses de retraite.
            interpolateur_B : 
                une fonction. 
                L'interpolateur de la part des revenus d'activités bruts dans le PIB. 
            interpolateur_NR : 
                une fonction. 
                L'interpolateur du nombre de retraités. 
            interpolateur_G : 
                une fonction. 
                L'interpolateur de l'effectif moyen d'une génération 
                arrivant aux âges de la retraite. 
            interpolateur_A : 
                une fonction. 
                L'interpolateur de l'âge effectif moyen de départ en retraite. 

        Exemple :
            modele = FonctionPension(simulateur, 2050)
            S = 0.0
            D = 0.14
            As = 63.0
            F = 0.5
            TauC = 7.0
            X = ot.Point([S, D, As, F, TauC])
            Y = modele(X)
        """
        super(FonctionPension, self).__init__(5, 1)
        # Attributs
        self.simulateur = simulateur
        self.annee = annee
        self.verbose = verbose
        # Configuration de la fonction
        self.setInputDescription(["S", "D", "As", "F", "TauC"])
        self.setOutputDescription(["P"])
        # Calcul des interpolateurs
        scenario_central = self.simulateur.scenario_central
        scenario_optimiste = self.simulateur.scenario_optimiste
        scenario_pessimiste = self.simulateur.scenario_pessimiste
        # Table des taux de chômages
        table_TauC = [self.simulateur.scenarios_chomage[scenario_optimiste], \
                      self.simulateur.scenarios_chomage[scenario_central], \
                      self.simulateur.scenarios_chomage[scenario_pessimiste]]
        # Table de NC
        table_NC = [self.simulateur.NC[scenario_optimiste][self.annee], \
                    self.simulateur.NC[scenario_central][self.annee], \
                    self.simulateur.NC[scenario_pessimiste][self.annee]]
        self.interpolateur_NC = sp.interpolate.interp1d(table_TauC, table_NC)
        # Table de dP
        table_dP = [self.simulateur.dP[scenario_optimiste][self.annee], \
                    self.simulateur.dP[scenario_central][self.annee], \
                    self.simulateur.dP[scenario_pessimiste][self.annee]]
        self.interpolateur_dP = sp.interpolate.interp1d(table_TauC, table_dP)
        # Table de B
        table_B = [self.simulateur.B[scenario_optimiste][self.annee], \
                   self.simulateur.B[scenario_central][self.annee], \
                   self.simulateur.B[scenario_pessimiste][self.annee]]
        self.interpolateur_B = sp.interpolate.interp1d(table_TauC, table_B)
        # Table de NR
        table_NR = [self.simulateur.NR[scenario_optimiste][self.annee], \
                    self.simulateur.NR[scenario_central][self.annee], \
                    self.simulateur.NR[scenario_pessimiste][self.annee]]
        self.interpolateur_NR = sp.interpolate.interp1d(table_TauC, table_NR)
        # Table de G
        table_G = [self.simulateur.G[scenario_optimiste][self.annee], \
                   self.simulateur.G[scenario_central][self.annee], \
                   self.simulateur.G[scenario_pessimiste][self.annee]]
        self.interpolateur_G = sp.interpolate.interp1d(table_TauC, table_G)
        # Table de A
        table_A = [self.simulateur.A[scenario_optimiste][self.annee], \
                   self.simulateur.A[scenario_central][self.annee], \
                   self.simulateur.A[scenario_pessimiste][self.annee]]
        self.interpolateur_A = sp.interpolate.interp1d(table_TauC, table_A)
        return

    def _exec(self, X):
        """
        Calcule la pension en fonction du solde S, 
        des dépenses D et de l'âge de départ en retraite As.
        
        Paramètres
        S : le solde financier
        D : le montant des dépenses de retraite en part de PIB
        As : l'âge de départ à la retraite de l'utilisateur
        F : facteur d'élasticité (par défaut, F = 0.5)
        TauC : taux de chômage (%)
        
        Description
        Retourne 
        P : niveau des pensions par rapport aux salaires
        
        Si l’utilisateur renseigne un âge effectif moyen de départ 
        à la retraite plus élevé que celui
        qui découle de l’évolution spontanée à législation inchangée 
        (c’est-à-dire si As > A), le nombre de
        retraités est donc diminué de G x (As – A) et le nombre 
        de cotisants augmenté de F x G x (As – A).
        
        Le calcul utilise la variable NC :
        NC : Nombre de personnes en emploi (ou nombre de cotisants)
    
        B : part des revenus d'activités bruts dans le PIB
        G : Effectif moyen d'une génération arrivant aux âges de la retraite
        A : âge moyen de départ à la retraite du COR
        NR : Nombre de retraités de droit direct (tous régimes confondus)
        dP : Autres dépenses de retraite rapportées au nombre de retraités 
             de droit direct en % du revenu d'activités brut moyen
        """
        S, D, As, F, TauC = X
        # Paramètres
        # Influence du taux de chômage
        G = self.interpolateur_G(TauC)
        A = self.interpolateur_A(TauC)
        NC = self.interpolateur_NC(TauC)
        dP = self.interpolateur_dP(TauC)
        B = self.interpolateur_B(TauC)
        NR = self.interpolateur_NR(TauC)
        # Coeur du modèle
        T = (S + D) / B
        g = G * (As - A)
        K = (NR - g) / (NC + F * g)
        P = (T - S / B) / K - dP
        # Affichage
        if self.verbose:
            print("G=", G)
            print("A=", A)
            print("NC =", NC)
            print("dP =", dP)
            print("B=", B)
            print("NR=", NR)
            print("T =", T)
            print("g =", g)
            print("K =", K)
            print("P =", P)
        # Sortie
        Y = [P]
        return Y



"""retraites module."""
from .SimulateurRetraites import SimulateurRetraites
from .SimulateurAnalyse import SimulateurAnalyse
from .EtudeImpact import EtudeImpact
from .FonctionPension import FonctionPension
from .ModelePensionProbabiliste import ModelePensionProbabiliste

__all__ = ['SimulateurRetraites', 'SimulateurAnalyse', 'EtudeImpact', \
           'FonctionPension', 'ModelePensionProbabiliste']
__version__ = '1.0'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import openturns as ot
from retraites.FonctionPension import FonctionPension

class ModelePensionProbabiliste():
    def InterpoleAge(annee, annee_courante, annee_horizon, 
                   age_courant, age_horizon):
        """
        Retourne l'âge pour l'année "annee" par interpolation linéaire 
        entre l'âge courant et l'âge à l'horizon. 
        
        Paramètres :
            annee : un flottant, l'année où calculer l'âge
            annee_courante : un flottant, l'année d'aujourd'hui
            annee_horizon : un flottant, l'année finale de la simulation dans le futur
            age_courant : un flottant, l'âge de départ en retraite aujourd'hui
            age_horizon : un flottant, l'âge de départ en retraite à la fin de la simulation
        
        Description
        Utilise une interpolation linéaire entre les deux points suivants :
        * (annee_courante, age_courant)
        * (annee_horizon, age_horizon)
        """
        age = (annee - annee_courante) / (annee_horizon - annee_courante) * age_horizon + \
              (annee_horizon - annee) / (annee_horizon - annee_courante) * age_courant
        return age

    def __init__(self, simulateur, annee, S, D, 
                 ageMin = 62.0, ageMax = 66.0, 
                 FMin = 0.25, FMax = 0.75, 
                 tauxChomageMin = 4.5, tauxChomageMax = 10.0,
                 bornesAgeConstant = True):
        """
        Crée un modèle de pension probabiliste.
        
        Paramètres :
            simulateur : un SimulateurRetraite
            annee : un flottant, l'année de calcul de P
            S : un flottant, le solde financier en part de PIB
            D : un flottant positif, le montant des dépenses de retraites 
            en part de PIB
            ageMin : un flottant, l'âge minimum
            ageMax : un flottant, l'âge maximum
            FMin : un flottant dans [0, 1], le facteur de report 
            de l'âge de départ en retraite minimum
            FMax : un flottant dans [0, 1], le facteur de report 
            de l'âge de départ en retraite maximum
            tauxChomageMin : un flottant positif, le taux de chômage 
            minimum
            tauxChomageMax : un flottant positif, le taux de chômage 
            maximum
            bornesAgeConstant : un booléen. Si True, alors utilise les bornes 
            ageMin et ageMax quelque soit l'année. 
            Sinon, utilise un âge situé entre l'âge du COR et l'âge de 
            l'étude d'impact.
            
        Attributs :
            fonction : 
                Une ot.Function. 
                Le modèle. 
        
            inputDistribution : 
                Une ot.Distribution. 
                La distribution des variables d'entrée. 
                
            ageMin :
                Un flottant. 
                La borne inférieure de la distribution de l'âge moyen 
                effectif de départ en retraite pour l'année. 
                
            ageMax :
                Un flottant. 
                La borne supérieure de la distribution de l'âge moyen 
                effectif de départ en retraite pour l'année. 
                
        Description :
            Crée un modèle de pension probabiliste pour le 
            ratio (pension moyenne) / (salaire moyen).
            
            Les entrées du modèle sont "As", "F", "TauC" 
            et la sortie est "P". 
            
            Les paramètres S et D sont fixés par le constructeur 
            de la classe au moment de la création de l'objet. 
            
            * S : le solde financier du système de retraites (% PIB)
            * D : le montant des dépenses (% PIB)
            * As : l'âge moyen de départ à la retraite défini par l'utilisateur
            * F  : facteur d'élasticité de report de l'âge de départ 
                (par exemple F=0.5)
            * TauC : le taux de chômage (par exemple TauC = 4.5)

            Les marginales du vecteur aléatoire sont indépendantes. 
            
            * F = ot.Uniform(FMin, FMax)
            * TauC = ot.Uniform(tauxChomageMin, tauxChomageMax)

            Si bornesAgeConstant est vrai, alors 
            As = ot.Uniform(ageMin, ageMax)
            
            Sinon, alors l'âge suit une distribution qui dépend de l'année 
            et dont les bornes sont entre l'âge du COR et l'âge de l'étude 
            d'impact. 

            * Avant 2020, la distribution de l'âge est un Dirac centré 
            sur l'âge du COR. 

            * De 2020 à 2038, l'âge du COR et l'âge de l'étude d'impact sont les mêmes.
            La distribution est un Dirac. 

            * De 2038 à 2044, l'âge de l'étude d'impact est inférieur à celui du COR. 
            L'étude d'impact prévoit une avance de l'âge de départ à la retraite sur cette période. 
            La distribution est uniforme. 
            De 2044 à 2070, l'âge de l'étude d'impact est supérieur à celui du COR. 
            L'étude d'impact prévoit un recul de l'âge de départ à la retraite sur cette période.
            La distribution est uniforme. 

        Exemple :
            S = 0.0
            D = 0.14
            annee = 2050
            modele = ModelePensionProbabiliste(simulateur, annee, S, D)
            fonction = modele.getFonction()
            inputDistribution = modele.getInputDistribution()
        """
        # Crée le modèle de pension complet : entrées = (S, D, As, F, TauC)
        modelePension = ot.Function(FonctionPension(simulateur, annee))
        # Crée le modèle réduit à partir du modèle complet : entrées = (As, F, TauC)
        indices = ot.Indices([0, 1])
        referencePoint = ot.Point([S, D])
        self.fonction = ot.ParametricFunction(modelePension, indices, 
                                              referencePoint)
        # Distribution
        self._calculeAge(simulateur, annee, ageMin, ageMax, bornesAgeConstant)
        if self.ageMin == self.ageMax:
            As = ot.Dirac(self.ageMin)
        else:
            As = ot.Uniform(self.ageMin, self.ageMax)
        F = ot.Uniform(FMin, FMax)
        TauC = ot.Uniform(tauxChomageMin, tauxChomageMax)
        self.inputDistribution = ot.ComposedDistribution([As, F, TauC])
        self.inputDistribution.setDescription(["As", "F", "TauC"])
        return

    def getFonction(self):
        """
        Retourne la fonction du modèle physique.
        """
        return self.fonction
    
    def getInputDistribution(self):
        """
        Retourne la distribution du modèle.
        """
        return self.inputDistribution
    
    def _calculeAge(self, simulateur, annee, ageMin, ageMax, 
                    bornesAgeConstant):
        """
        Calcule les bornes de l'âge en fonction de l'option bornesAgeConstant.
        
        Si bornesAgeConstant est True, alors utilise les arguments 
        ageMin et ageMax.
        Sinon, 
        * si l'année est inférieure à l'année courante, utilise l'âge du COR,
        * sinon, réalise une interpolation linéaire entre les paramètres 
        de l'étude d'impact (borne supérieure) et les paramètres du COR (borne
        inférieure).
        """
        if bornesAgeConstant:
            self.ageMin = ageMin
            self.ageMax = ageMax
        else:
            # Pour l'âge de départ en retraite
            analyse_COR = simulateur.pilotageCOR()
            if annee <= simulateur.annee_courante:
                ageCOR = analyse_COR.A[simulateur.scenario_central][annee]
                self.ageMin = ageCOR
                self.ageMax = ageCOR
            else:
                ageCOR_courant = analyse_COR.A[simulateur.scenario_central][simulateur.annee_courante]
                ageMin_horizon = ageMin
                ageMax_horizon = ageMax
                self.ageMin = ModelePensionProbabiliste.InterpoleAge(annee, 
                                           simulateur.annee_courante, simulateur.horizon, 
                                           ageCOR_courant, ageMin_horizon)
                self.ageMax = ModelePensionProbabiliste.InterpoleAge(annee, 
                                                simulateur.annee_courante, simulateur.horizon, 
                                                ageCOR_courant, ageMax_horizon)
        return
        
    


# coding: utf8
"""
Setup script for retraites
==========================

This script allows to install retraites within the Python environment.

Usage
-----
::

    python setup.py install

"""
from setuptools import (setup, find_packages)

install_requires = ['numpy',
                    'matplotlib',
                    'scipy',
                    'openturns',
                    ]
extras_require = {'doc': ['jupyter', 'jupyter_client']}

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='retraites',
    keywords=("graphics"),
    version='0.0.1',
    packages=find_packages(exclude=['doc']),
    install_requires=install_requires,
    extras_require=extras_require,
    description="retraites: Simulateur d'équilibre financier du système de retraites",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=['Development Status :: 5 - Production/Stable',
                 'Environment :: Console',
                 'License :: OSI Approved',
                 'Intended Audience :: Science/Research',
                 'Intended Audience :: Developers',
                 'Natural Language :: English',
                 'Operating System :: Unix',
                 'Programming Language :: Python :: 3',
                 'Topic :: Documentation :: Sphinx',
                 'Topic :: Software Development',
                 'Topic :: Scientific/Engineering',
                 ],
    include_package_data=True,
    zip_safe=False,
    license="GPL",
    url="https://github.com/brunoscherrer/retraites",
)



#!/usr/bin/python
# coding:utf-8
"""
Classe de gestion d'un simulateur de retraites.
"""

from copy import deepcopy
import json
from retraites.SimulateurAnalyse import SimulateurAnalyse
import pylab as pl
import os
import retraites
import scipy.optimize as spo

class SimulateurRetraites:

    def __init__(self, json_filename = None):
        """
        Crée un simulateur à partir d'un fichier d'hypothèses JSON.
        
        Paramètres :
            json_filename : une chaîne de caractère, le nom du fichier JSON 
            contenant les hypothèses
            (par défaut, charge le fichier "fileProjection.json" fournit par le 
            module)
            pilotage : un entier, la stratégie de pilotage (par défaut, celle du COR)
            
        Attributs :
            annee_courante : 
                Un entier. 
                L'année correspondant à la date d'aujourd'hui.

            horizon : 
                Un entier. 
                La dernière année du calcul. 
    
            annees : 
                Une liste d'entiers.
                La liste des années sur lesquelles on fait les calculs. 
                Chaque année de cette liste est inférieure à l'année de l'horizon.
    
            annees_futures :
                Une liste d'entiers.
                La liste des années sur lesquelles on peut changer 
                quelque chose.
    
            annees_standard :
                Une liste d'entiers.
                La liste d'une sélection des années futures standard dans les calculs simplifiés. 
    
            annees_EV :
                Une liste d'entiers.
                La liste des années de naissance pour lesquelles on a l'espérance de vie. 
    
            scenarios :
                Une liste d'entiers.
                la liste des scénarios considérés. 
                Ces scénarios sont des indices dans les tables de scénarios de 
                chomage, de croissance ainsi que les labels. 
    
            scenario_central :
                Un entier. 
                L'indice du scénario central, 
                +1,3%/an, Chômage: 7%. 
    
            scenario_pessimiste :
                L'indice du scénario pessimiste +1%/an, Chômage: 10%. 
        
            scenario_optimiste :
                L'indice du scénario optimiste : +1,8%/an, Chômage: 4.5%.
    
            scenarios_croissance :
                Une liste de flottants. 
                La liste des taux de croissance pour chaque scénario 
                de la liste retournée par getScenarios().
    
            scenarios_chomage :
                Une liste de flottants. 
                La liste des taux de chomage pour chaque scénario 
                de la liste retournée par getScenarios().
    
            scenarios_labels :
                La liste de chaîne de caractère décrivant les 
                scénarios pour chaque scénario 
                de la liste retournée par getScenarios().
    
            scenarios_labels_courts :
                La liste de chaîne de caractère courtes décrivant les 
                scénarios pour chaque scénario 
                de la liste retournée par getScenarios().
                
            T : 
                Un dictionnaire représentant une trajectoire. 
                Le taux de cotisations retraites
            
            P : 
                Un dictionnaire représentant une trajectoire. 
                Le niveau moyen brut des pensions par rapport au 
                niveau moyen brut des salaires
            
            A : 
                Un dictionnaire représentant une trajectoire. 
                l'âge effectif moyen de départ en retraite
                
            G : 
                Un dictionnaire représentant une trajectoire. 
                Effectif moyen d'une génération arrivant aux âges 
                de la retraite
                
            NR  : 
                Un dictionnaire représentant une trajectoire. 
                Nombre de retraités de droit direct (tous régimes confondus)
                
            NC :
                Un dictionnaire représentant une trajectoire. 
                Nombre de personnes en emploi (ou nombre de cotisants) 
        
            TCR :
                Un dictionnaire représentant une trajectoire. 
                Taux des prélèvements sociaux sur les pensions de retraite
                Son nom est TPR dans le composant, TCR dans le fichier json
        
            TCS :
                Un dictionnaire représentant une trajectoire. 
                Taux des prélèvements sociaux sur les salaires et 
                revenus d'activité ;
                Son nom est TPR dans le composant, TCR dans le fichier json
        
            CNV :
                Un dictionnaire représentant une trajectoire. 
                Coefficient pour passer du ratio "pensions/salaire moyen" 
                au ratio "niveau de vie/salaire moyen"
        
            dP :
                Un dictionnaire représentant une trajectoire. 
                Autres dépenses de retraite rapportées au nombre de 
                retraités de droit direct en % du revenu d'activités brut moyen
        
            B :
                Un dictionnaire représentant une trajectoire. 
                part des revenus d'activités bruts dans le PIB 
        
            EV :
                Un dictionnaire représentant une trajectoire. 
                Espérance de vie à 60 ans par génération

            liste_variables :
                Une liste de chaînes de caractères. 
                La liste des variables du modèle : B, NR, etc...
                
            liste_legendes
                Une liste de chaînes de caractères. 
                La liste des légendes pour chaque variable dans liste_variables

            labels_is_long :
                Un booléen. 
                True, si on utilise les labels longs dans les graphiques
        
            yaxis_lim :
                Un dictionnaire. 
                Les plages min et max pour l'axe des ordonnées 
                des variables en sortie du simulateur. 

            dir_image :
                Une chaîne de caractère. 
                Le répertoire de sauvegarde des images. 
                Par défaut, le répertoire courant. 

            ext_image :
                Une liste de chaînes de caractères. 
                Les types de fichier à générer par la méthode sauveFigure. 
                
            affiche_quand_ecrit :
                Un booléen.
                Si True, alors affiche un message quand la méthode sauveFigure 
                écrit un fichier.

            rechercheAgeBornes :
                Une liste de flottants.
                Les bornes de recherches pour l'inversion de l'âge 
                en fonction du ratio de durée de vie en retraite. 

            rechercheAgeRTol :
                Un flottant.
                La tolérance relative sur l'âge pour l'inversion de l'âge 
                en fonction du ratio de durée de vie en retraite. 
        
        Description :
            Plusieurs stratégies de pilotage peuvent être utilisées :
            1 pilotageCOR, avec les paramètres du COR
            1 pilotageParPensionAgeCotisations
            2 pilotageParSoldePensionAge
            3 pilotageParSoldePensionCotisations
            4 pilotageParSoldeAgeCotisations
            5 pilotageParSoldeAgeDepenses
            6 pilotageParSoldePensionDepenses
            7 pilotageParPensionCotisationsDepenses
            8 pilotageParAgeCotisationsDepenses
            9 pilotageParAgeEtNiveauDeVie (sous-entendu et par solde financier)
            10 pilotageParNiveauDeVieEtCotisations (sous-entendu et par solde financier)
            
            Les scénarios sont numérotés de 1 à 6 dans l'attribut "scenarios"
            (contrairement à l'usage Python ordinaire qui voudrait plutôt que 
            l'indice aille de 0 à 5). 

        Exemple :
            simulateur = SimulateurRetraites()
            simulateur.dessineConjoncture()
            simulateur.dessineLegende()
            
            simulateur = SimulateurRetraites()
            analyse = simulateur.pilotageCOR()
            analyse.dessineSimulation()
            analyse.dessineLegende()
        """
        
        if (json_filename is None):
            # Loading default JSON data
            json_filename = os.path.join(retraites.__path__[0], "fileProjection.json")
            
        # initialisations diverses
        # chargement des donnees du COR pour les 6 scenarios
        
        # Lit les hypothèses de calcul dans le fichier JSON
        json_file = open(json_filename)
        self.data = json.load(json_file)
        json_file.close()
        
        # Paramètres constants
        self.annee_courante = 2020  # Annee correspondant à la date d'aujourd'hui
        self.horizon = 2070         # Dernière année du calcul
        self.annees_futures=range(self.annee_courante, self.horizon+1) # annees sur lesquelles on peut changer qqch
        self.annees=range(2005, self.horizon+1)         # annees sur lesquelles on fait les calculs
        self.annees_standard=[2020, 2025, 2030, 2040, 2050, 2060, 2070] # Années standard dans les calculs simplifiés
        self.annees_EV=range(1930,2011)                 # annees sur lesquelles on a l'espérance de vie

        # Scénarios
        self.scenarios = range(1,7)  # Scenarios considérés
        self.scenario_central = 3    # central    : +1,3%/an, Chômage: 7%
        self.scenario_pessimiste = 6 # pessimiste :   +1%/an, Chômage: 10%
        self.scenario_optimiste = 5  # optimiste  : +1,8%/an, Chômage: 4.5%
        # Taux de croissance pour chaque scénario
        self.scenarios_croissance = [0.0, 1.8, 1.5, 1.3, 1.0, 1.8, 1.0]
        # Taux de chomage pour chaque scénario
        self.scenarios_chomage = [0.0, 7.0, 7.0, 7.0, 7.0, 4.5, 10.0]
        # Graphiques
        self.scenarios_labels=["Scénario inexistant", 
                               "Hausse des salaires: +1,8%/an, Taux de chômage: 7%",
                               "Hausse des salaires: +1,5%/an, Taux de chômage: 7%",
                               "Hausse des salaires: +1,3%/an, Taux de chômage: 7%",
                               "Hausse des salaires: +1%/an, Taux de chômage: 7%",
                               "Hausse des salaires: +1,8%/an, Taux de chômage: 4.5%",
                               "Hausse des salaires: +1%/an, Taux de chômage: 10%"]
        self.scenarios_labels_courts=["Scénario inexistant", 
                                      "+1,8%/an, Chômage: 7%",
                                      "+1,5%/an, Chômage: 7%",
                                      "+1,3%/an, Chômage: 7%",
                                      "+1%/an, Chômage: 7%",
                                      "+1,8%/an, Chômage: 4.5%",
                                      "+1%/an, Chômage: 10%"]

        # Extrait les variables depuis les données
        self.T = self.get('T')
        self.P = self.get('P')
        self.A = self.get('A')
        self.G = self.get('G')
        self.NR = self.get('NR')
        self.NC = self.get('NC')
        self.TCR = self.get('TCR') # Son nom est TPR dans le composant, TCR dans le fichier json
        self.TCS = self.get('TCS') # Son nom est TCS dans le composant, TCS dans le fichier json
        self.CNV = self.get('CNV')
        self.dP = self.get('dP')
        self.B = self.get('B')
        self.EV = self.get('EV')

        self.liste_variables = ["B","NR","NC","G","dP","TPR","TPS","CNV","EV"]
        self.liste_legendes=[u"B: Part des revenus d'activité bruts dans le PIB",
           u"NR: Nombre de retraités",
           u"NC: Nombre de cotisants",
           u"G: Effectif d'une generation arrivant à l'âge de la retraite",
           u"dP: Autres dépenses de retraites",
           u"TPR: Taux de prélèvement sur les retraites",
           u"TPS: Taux de prélèvement sur les salaires",
           u"CNV: (niveau de vie)/[(pension moy))/(salaire moy)]",
           u"EV: Espérance de vie à 60 ans"
        ]

        self.labels_is_long = True # True, si on utilise les labels longs
        
        # Configure les plages min et max pour l'axe des ordonnées 
        # des variables standard en sortie du simulateur
        self.yaxis_lim = dict()
        self.yaxis_lim["RNV"] = [60.0,120.0]
        self.yaxis_lim["REV"] = [20.0,40.0]

        self.ext_image=["png","pdf"]   # types de fichier à générer
        
        # Le répertoire de sauvegarde des images
        self.dir_image = "."

        self.affiche_quand_ecrit = True # Affiche un message quand on écrit un fichier

        # Paramètres pour l'algorithme d'inversion de la durée de vie en retraite
        # Bornes de recherche de l'âge
        self.rechercheAgeBornes = [60.0, 70.0]
        # Tolérance relative sur l'âge
        self.rechercheAgeRTol = 1.e-3
        return None

    def pilotageCOR(self):
        """
        pilotage 1 : statu quo du COR
        
        Description :
            Retourne un objet de type SimulateurAnalyse.

        Exemple :
            simulateur = SimulateurRetraites()
            simulateur.pilotageCOR()
        """
        S, RNV, REV, Depenses = self._calcule_S_RNV_REV(self.T, self.P, self.A)
        resultat = self._creerAnalyse(self.T, self.P, self.A, S, RNV, REV, Depenses)
        return resultat
    
    def _creerAnalyse(self, Ts, Ps, As, Ss, RNVs, REVs, Depenses):
        """
        Retourne une analyse en fonction des objets essentiels. 
        """
        PIB = self._genereTrajectoirePIB()
        PensionBrut = self._calculePensionAnnuelleDroitDirect(PIB, As)
        resultat = SimulateurAnalyse(Ts, Ps, As, Ss, RNVs, REVs, Depenses, \
                                     PIB, PensionBrut, \
                                     self.scenarios, self.annees_EV, self.annees, self.annees_standard, \
                                     self.scenarios_labels, self.scenarios_labels_courts, 
                                     self.dir_image, self.ext_image)
        return resultat
    
    def pilotageParPensionAgeCotisations(self, Pcible=None, Acible=None, Tcible=None):
        """
        pilotage 1 : imposer 1) le niveau des pensions par rapport aux salaires
        2) l'âge de départ à la retraite
        3) le taux de cotisations
            
        Paramètres :
            Pcible : le niveau de pension des retraites par rapport aux actifs
            Acible : l'âge de départ à la retraite
            Tcible : le taux de cotisations

        Description :
            Retourne un objet de type SimulateurAnalyse.
            * Si la valeur cible n'est pas donnée, utilise par défaut la trajectoire du COR.
            * Si la valeur cible donnée est un flottant, utilise la trajectoire du 
            COR pour les années passées et cette valeur pour les années futures. 
            * Si la valeur cible donnée est un dictionnaire, considère que c'est 
            une trajectoire et utilise cette trajectoire. 

        Exemple :
            simulateur = SimulateurRetraites()
            simulateur.pilotageParPensionAgeCotisations(Pcible = 0.5)
            simulateur.pilotageParPensionAgeCotisations(Acible = 62.0)
            simulateur.pilotageParPensionAgeCotisations(Tcible = 0.28)
            simulateur.pilotageParPensionAgeCotisations(Pcible = 0.5, Acible = 62.0)
            simulateur.pilotageParPensionAgeCotisations(Pcible = 0.5, Acible = 62.0, Tcible = 0.28)
            
            # Conserve le niveau de pension actuel
            s = 3 # Scénario central
            Pcible = simulateur.P[s][2020]
            simulateur.pilotageParPensionAgeCotisations(Pcible = Pcible)
        """
        # Génère les trajectoires en fonction des paramètres
        Ps = self.genereTrajectoire("P", Pcible)
        As = self.genereTrajectoire("A", Acible)
        Ts = self.genereTrajectoire("T", Tcible)
        # Simule 
        S, RNV, REV, Depenses = self._calcule_S_RNV_REV(Ts,Ps,As)
        resultat = self._creerAnalyse(Ts, Ps, As, S, RNV, REV, Depenses)
        return resultat

    def pilotageParSoldePensionAge(self, Scible=None, Pcible=None, Acible=None):
        """
        pilotage 2 : imposer 1) le bilan financer
        2) le niveau des pensions par rapport aux salaires
        3) l'âge de départ à la retraite
            
        Paramètres :
            Scible : la situation financière en % de PIB
            Pcible : le niveau de pension des retraites par rapport aux actifs
            Acible : l'âge de départ à la retraite
            
        Description :
            Retourne un objet de type SimulateurAnalyse.
            * Si la valeur cible n'est pas donnée, utilise par défaut la trajectoire du COR.
            * Si la valeur cible donnée est un flottant, utilise la trajectoire du 
            COR pour les années passées et cette valeur pour les années futures. 
            * Si la valeur cible donnée est un dictionnaire, considère que c'est 
            une trajectoire et utilise cette trajectoire. 

        Exemple :
            simulateur = SimulateurRetraites()
            simulateur.pilotageParSoldePensionAge(Scible = 0.0)
        """
        # Génère les trajectoires en fonction des paramètres
        Ss = self.genereTrajectoire("S", Scible)
        Ps = self.genereTrajectoire("P", Pcible)
        As = self.genereTrajectoire("A", Acible)
        # Calcule le pilotage
        Ts, Ps, As = self._calcule_fixant_Ss_Ps_As(Ss, Ps, As)
        # Simule 
        S, RNV, REV, Depenses = self._calcule_S_RNV_REV(Ts,Ps,As)
        resultat = self._creerAnalyse(Ts, Ps, As, S, RNV, REV, Depenses)
        return resultat

    def pilotageParSoldePensionCotisations(self, Scible=None, Pcible=None, Tcible=None):
        """
        pilotage 3 : imposer 1) le bilan financer
        2) le niveau des pensions par rapport aux salaires
        3) le taux de cotisations
            
        Paramètres :
            Scible : la situation financière en % de PIB
            Pcible : le niveau de pension des retraites par rapport aux actifs
            Tcible : le taux de cotisations
            
        Description :
            Retourne un objet de type SimulateurAnalyse.
            * Si la valeur cible n'est pas donnée, utilise par défaut la trajectoire du COR.
            * Si la valeur cible donnée est un flottant, utilise la trajectoire du 
            COR pour les années passées et cette valeur pour les années futures. 
            * Si la valeur cible donnée est un dictionnaire, considère que c'est 
            une trajectoire et utilise cette trajectoire. 

        Exemple :
            simulateur = SimulateurRetraites()
            simulateur.pilotageParSoldePensionCotisations(Scible = 0.0)
        """
        # Génère les trajectoires en fonction des paramètres
        Ss = self.genereTrajectoire("S", Scible)
        Ps = self.genereTrajectoire("P", Pcible)
        Ts = self.genereTrajectoire("T", Tcible)
        # Calcule le pilotage
        Ts, Ps, As = self._calcule_fixant_Ss_Ps_Ts(Ss, Ps, Ts)
        # Simule 
        S, RNV, REV, Depenses = self._calcule_S_RNV_REV(Ts,Ps,As)
        resultat = self._creerAnalyse(Ts, Ps, As, S, RNV, REV, Depenses)
        return resultat 

    def pilotageParSoldeAgeCotisations(self, Scible=None, Acible=None, Tcible=None):
        """
        pilotage 4 : imposer 1) le bilan financer
        2) l'âge de départ à la retraite
        3) le taux de cotisations
            
        Paramètres :
            Scible : la situation financière en % de PIB
            Acible : l'âge de départ à la retraite
            Tcible : le taux de cotisations
            
        Description :
            Retourne un objet de type SimulateurAnalyse.
            * Si la valeur cible n'est pas donnée, utilise par défaut la trajectoire du COR.
            * Si la valeur cible donnée est un flottant, utilise la trajectoire du 
            COR pour les années passées et cette valeur pour les années futures. 
            * Si la valeur cible donnée est un dictionnaire, considère que c'est 
            une trajectoire et utilise cette trajectoire. 

        Exemple :
            simulateur = SimulateurRetraites()
            simulateur.pilotageParSoldeAgeCotisations(Scible = 0.0)
        """
        # Génère les trajectoires en fonction des paramètres
        Ss = self.genereTrajectoire("S", Scible)
        As = self.genereTrajectoire("A", Acible)
        Ts = self.genereTrajectoire("T", Tcible)
        # Calcule le pilotage
        Ts, Ps, As = self._calcule_fixant_Ss_As_Ts(Ss, As, Ts)
        # Simule 
        S, RNV, REV, Depenses = self._calcule_S_RNV_REV(Ts,Ps,As)
        resultat = self._creerAnalyse(Ts, Ps, As, S, RNV, REV, Depenses)
        return resultat 

    def pilotageParSoldeAgeDepenses(self, Scible=None, Acible=None, Dcible=None):
        """
        pilotage 5 : imposer 1) le bilan financer
        2) l'âge de départ à la retraite
        3) le niveau de dépenses
            
        Paramètres :
            Scible : la situation financière en % de PIB
            Acible : l'âge de départ à la retraite
            Dcible : le niveau de dépenses
            
        Description :
            Retourne un objet de type SimulateurAnalyse.
            * Si la valeur cible n'est pas donnée, utilise par défaut la trajectoire du COR.
            * Si la valeur cible donnée est un flottant, utilise la trajectoire du 
            COR pour les années passées et cette valeur pour les années futures. 
            * Si la valeur cible donnée est un dictionnaire, considère que c'est 
            une trajectoire et utilise cette trajectoire. 

        Exemple :
            simulateur = SimulateurRetraites()
            simulateur.pilotageParSoldeAgeDepenses(Scible = 0.0)
        """
        # Génère les trajectoires en fonction des paramètres
        Ss = self.genereTrajectoire("S", Scible)
        As = self.genereTrajectoire("A", Acible)
        Ds = self.genereTrajectoire("Depenses", Dcible)
        # Calcule le pilotage
        Ts, Ps, As = self._calcule_fixant_Ss_As_Ds(Ss, As, Ds)
        # Simule 
        S, RNV, REV, Depenses = self._calcule_S_RNV_REV(Ts,Ps,As)
        resultat = self._creerAnalyse(Ts, Ps, As, S, RNV, REV, Depenses)
        return resultat 
    
    def pilotageParSoldePensionDepenses(self, Scible=None, Pcible=None, Dcible=None):
        """
        pilotage 6 : imposer 1) le bilan financer
        2) le niveau des pensions par rapport aux salaires
        3) le niveau de dépenses
            
        Paramètres :
            Scible : la situation financière en % de PIB
            Pcible : le niveau de pension des retraites par rapport aux actifs
            Dcible : le niveau de dépenses
            
        Description :
            Retourne un objet de type SimulateurAnalyse.
            * Si la valeur cible n'est pas donnée, utilise par défaut la trajectoire du COR.
            * Si la valeur cible donnée est un flottant, utilise la trajectoire du 
            COR pour les années passées et cette valeur pour les années futures. 
            * Si la valeur cible donnée est un dictionnaire, considère que c'est 
            une trajectoire et utilise cette trajectoire. 

        Exemple :
            simulateur = SimulateurRetraites()
            simulateur.pilotageParSoldePensionDepenses(Scible = 0.0)
        """
        # Génère les trajectoires en fonction des paramètres
        Ss = self.genereTrajectoire("S", Scible)
        Ps = self.genereTrajectoire("P", Pcible)
        Ds = self.genereTrajectoire("Depenses", Dcible)
        # Calcule le pilotage
        Ts, Ps, As = self._calcule_fixant_Ss_Ps_Ds(Ss, Ps, Ds)
        # Simule 
        S, RNV, REV, Depenses = self._calcule_S_RNV_REV(Ts,Ps,As)
        resultat = self._creerAnalyse(Ts, Ps, As, S, RNV, REV, Depenses)
        return resultat 

    def pilotageParPensionCotisationsDepenses(self, Pcible=None, Tcible=None, Dcible=None):
        """
        pilotage 7 : imposer 1) le niveau des pensions par rapport aux salaires
        2) le taux de cotisations
        3) le niveau de dépenses
            
        Paramètres :
            Pcible : le niveau de pension des retraites par rapport aux actifs
            Tcible : le taux de cotisations
            Dcible : le niveau de dépenses
            
        Description :
            Retourne un objet de type SimulateurAnalyse.
            * Si la valeur cible n'est pas donnée, utilise par défaut la trajectoire du COR.
            * Si la valeur cible donnée est un flottant, utilise la trajectoire du 
            COR pour les années passées et cette valeur pour les années futures. 
            * Si la valeur cible donnée est un dictionnaire, considère que c'est 
            une trajectoire et utilise cette trajectoire. 

        Exemple :
            simulateur = SimulateurRetraites()
            simulateur.pilotageParPensionCotisationsDepenses(Pcible = 0.5)
        """
        # Génère les trajectoires en fonction des paramètres
        Ps = self.genereTrajectoire("P", Pcible)
        Ts = self.genereTrajectoire("T", Tcible)
        Ds = self.genereTrajectoire("Depenses", Dcible)
        # Calcule le pilotage
        Ts, Ps, As = self._calcule_fixant_Ps_Ts_Ds(Ps, Ts, Ds)
        # Simule 
        S, RNV, REV, Depenses = self._calcule_S_RNV_REV(Ts,Ps,As)
        resultat = self._creerAnalyse(Ts, Ps, As, S, RNV, REV, Depenses)
        return resultat 

    def pilotageParAgeCotisationsDepenses(self, Acible=None, Tcible=None, Dcible=None):
        """
        pilotage 8 : imposer 1) l'âge de départ à la retraite
        2) le taux de cotisations
        3) le niveau de dépenses
            
        Paramètres :
            Acible : l'âge de départ à la retraite
            Tcible : le taux de cotisations
            Dcible : le niveau de dépenses
            
        Description :
            Retourne un objet de type SimulateurAnalyse.
            * Si la valeur cible n'est pas donnée, utilise par défaut la trajectoire du COR.
            * Si la valeur cible donnée est un flottant, utilise la trajectoire du 
            COR pour les années passées et cette valeur pour les années futures. 
            * Si la valeur cible donnée est un dictionnaire, considère que c'est 
            une trajectoire et utilise cette trajectoire. 

        Exemple :
            simulateur = SimulateurRetraites()
            simulateur.pilotageParAgeCotisationsDepenses(Acible = 62.0)
        """
        # Génère les trajectoires en fonction des paramètres
        As = self.genereTrajectoire("A", Acible)
        Ts = self.genereTrajectoire("T", Tcible)
        Ds = self.genereTrajectoire("Depenses", Dcible)
        # Calcule le pilotage
        Ts, Ps, As = self._calcule_fixant_As_Ts_Ds(As, Ts, Ds)
        # Simule 
        S, RNV, REV, Depenses = self._calcule_S_RNV_REV(Ts,Ps,As)
        resultat = self._creerAnalyse(Ts, Ps, As, S, RNV, REV, Depenses)
        return resultat 
    
    def pilotageParAgeEtNiveauDeVie(self, Acible=None, RNVcible=None, Scible=None):
        """
        pilotage 9 : imposer 1) l'âge de départ à la retraite,  
        2) le niveau de vie par rapport à l'ensemble de la population et 
        3) le bilan financier
        
        Paramètres :
            Acible : un flottant, l'âge de départ imposé
            RNVcible : un flottant positif, le niveau de vie des retraités par 
                rapport à l’ensemble de la population
            Scible : la situation financière en % de PIB
        
        Description :
            Retourne un objet de type SimulateurAnalyse.
            * Si la valeur cible n'est pas donnée, utilise par défaut la trajectoire du COR.
            * Si la valeur cible donnée est un flottant, utilise la trajectoire du 
            COR pour les années passées et cette valeur pour les années futures. 
            * Si la valeur cible donnée est un dictionnaire, considère que c'est 
            une trajectoire et utilise cette trajectoire. 

        Exemple :
            simulateur = SimulateurRetraites()
            simulateur.pilotageParAgeEtNiveauDeVie(RNVcible = 1.0)
        """
        # Génère les trajectoires en fonction des paramètres
        As = self.genereTrajectoire("A", Acible)
        RNVs = self.genereTrajectoire("RNV", RNVcible)
        Ss = self.genereTrajectoire("S", Scible)
        # Calcule le pilotage
        Ts, Ps, As = self._calcule_fixant_As_RNV_S(As, RNVs, Ss)
        # Simule 
        S, RNV, REV, Depenses = self._calcule_S_RNV_REV(Ts,Ps,As)
        resultat = self._creerAnalyse(Ts, Ps, As, S, RNV, REV, Depenses)
        return resultat
    
    def pilotageParNiveauDeVieEtCotisations(self, Tcible=None, RNVcible=None, Scible=None):
        """
        pilotage 10 : imposer 1) le taux de cotisations, 
        2) le niveau de vie par rapport à l'ensemble de la population et 
        3) le bilan financier
        
        Paramètres :
            Tcible : le taux de cotisations
            RNVcible : le niveau de vie des retraités par rapport à 
                l’ensemble de la population
            Scible : la situation financière en % de PIB
        
        Description :
            Retourne un objet de type SimulateurAnalyse.
            * Si la valeur cible n'est pas donnée, utilise par défaut la trajectoire du COR.
            * Si la valeur cible donnée est un flottant, utilise la trajectoire du 
            COR pour les années passées et cette valeur pour les années futures. 
            * Si la valeur cible donnée est un dictionnaire, considère que c'est 
            une trajectoire et utilise cette trajectoire.

        Exemple :
            simulateur = SimulateurRetraites()
            simulateur.pilotageParNiveauDeVieEtCotisations(RNVcible = 1.0)
        """
        # Génère les trajectoires en fonction des paramètres
        Ts = self.genereTrajectoire("T", Tcible)
        RNVs = self.genereTrajectoire("RNV", RNVcible)
        Ss = self.genereTrajectoire("S", Scible)
        # Calcule le pilotage
        Ts, Ps, As = self._calcule_fixant_Ts_RNV_S(Ts, RNVs, Ss)
        # Simule
        S, RNV, REV, Depenses = self._calcule_S_RNV_REV(Ts,Ps,As)
        resultat = self._creerAnalyse(Ts, Ps, As, S, RNV, REV, Depenses)
        return resultat
    
    def get(self, var):
        """
        Retourne une donnée du COR correspondant à un nom donné.
        
        Paramètres :
            var : une chaîne de caractère, la variable à extraire
            v : un dictionnaire, v[s][a] est la valeur de la variable 
            pour le scénario s à l'année a
        
        Exemple :
            T = simulateur.get("T")
        """
        if var=='EV':
            an=self.annees_EV
        else:
            an=self.annees

        v=dict()
        
        for s in self.scenarios:
            v[s]=dict()
            for a in an:
                v[s][a]=self.data[var][str(s)][str(a)]
                
        return v

    def _calcule_fixant_Ss_Ps_As(self, Ss, Ps, As):
        """
        Calcul à solde, pension et âge définis
        
        Paramètres :
            Ss : un dictionnaire, Ss[s][a] est le solde financier 
            du scénario s à l'année a
            Ps : un dictionnaire, Ps[s][a] est le montant de la pension 
            par rapport aux salaires du scénario s à l'année a
            As : un dictionnaire, As[s][a] est l'âge de départ à la retraite 
            du scénario s à l'année a
        """
    
        # Calcule Ts
        Ts = deepcopy(self.T)    
        for s in self.scenarios:    
            for a in self.annees_futures:
                GdA = self.G[s][a] * ( As[s][a] - self.A[s][a] )
                K = (self.NR[s][a] - GdA) / (self.NC[s][a] + 0.5*GdA)
                Ts[s][a] = Ss[s][a] / self.B[s][a] + K * (Ps[s][a] + self.dP[s][a])
                
        return Ts, Ps, As

    def _calcule_fixant_Ss_Ps_Ts(self, Ss, Ps, Ts):
        """
        Calcul à solde, pension et cotisations définis
        
        Paramètres :
            Ss : un dictionnaire, Ss[s][a] est le solde financier 
            du scénario s à l'année a
            Ps : un dictionnaire, Ps[s][a] est le montant de la pension 
            par rapport aux salaires du scénario s à l'année a
            Ts : un dictionnaire, Ts[s][a] est le taux de cotisations 
            du scénario s à l'année a
        """
    
        # Calcule As
        As = deepcopy(self.A)    
        for s in self.scenarios:    
            for a in self.annees_futures:
                K = (Ts[s][a] - Ss[s][a] / self.B[s][a]) / (Ps[s][a] + self.dP[s][a])
                As[s][a] = self.A[s][a] + (self.NR[s][a] - K * self.NC[s][a]) / (0.5*K + 1) / self.G[s][a]
                
        return Ts, Ps, As

    def _calcule_fixant_Ss_As_Ts(self, Ss, As, Ts):
        """
        Calcul à solde, âge et cotisations définis
        
        Paramètres :
            Ss : un dictionnaire, Ss[s][a] est le solde financier 
            du scénario s à l'année a
            As : un dictionnaire, As[s][a] est l'âge de départ à la retraite 
            du scénario s à l'année a
            Ts : un dictionnaire, Ts[s][a] est le taux de cotisations 
            du scénario s à l'année a
        """
    
        # Calcule Ps
        Ps = deepcopy(self.P)
        for s in self.scenarios:    
            for a in self.annees_futures:
                GdA = self.G[s][a] * ( As[s][a] - self.A[s][a] )
                K = (self.NR[s][a] - GdA) / (self.NC[s][a] + 0.5*GdA)
                Ps[s][a] = (Ts[s][a] - Ss[s][a] / self.B[s][a]) / K - self.dP[s][a]
                
        return Ts, Ps, As

    def _calcule_fixant_Ss_As_Ds(self, Ss, As, Ds):
        """
        Calcul à solde, âge et dépenses définis
        
        Paramètres :
            Ss : un dictionnaire, Ss[s][a] est le solde financier 
            du scénario s à l'année a
            As : un dictionnaire, As[s][a] est l'âge de départ à la retraite 
            du scénario s à l'année a
            Ds : un dictionnaire, Ds[s][a] est le niveau de dépenses 
            du scénario s à l'année a
        """
    
        # Calcule Ps et Ts
        Ps, Ts = deepcopy(self.P), deepcopy(self.T)
        for s in self.scenarios:    
            for a in self.annees_futures:
                Ts[s][a] = (Ss[s][a] + Ds[s][a]) / self.B[s][a]
                GdA = self.G[s][a] * ( As[s][a] - self.A[s][a] )
                K = (self.NR[s][a] - GdA) / (self.NC[s][a] + 0.5*GdA)
                Ps[s][a] = (Ts[s][a] - Ss[s][a] / self.B[s][a]) / K - self.dP[s][a]
                
        return Ts, Ps, As

    def _calcule_fixant_Ss_Ps_Ds(self, Ss, Ps, Ds):
        """
        Calcul à solde, pension et dépenses définis
        
        Paramètres :
            Ss : un dictionnaire, Ss[s][a] est le solde financier 
            du scénario s à l'année a
            Ps : un dictionnaire, Ps[s][a] est le montant de la pension 
            par rapport aux salaires du scénario s à l'année a
            Ds : un dictionnaire, Ds[s][a] est le niveau de dépenses 
            du scénario s à l'année a
        """
    
        # Calcule As et Ts
        As, Ts = deepcopy(self.A), deepcopy(self.T)
        for s in self.scenarios:    
            for a in self.annees_futures:
                K = Ds[s][a] / self.B[s][a] / (Ps[s][a] + self.dP[s][a])
                As[s][a] = self.A[s][a] + (self.NR[s][a] - K * self.NC[s][a]) / (0.5*K + 1) / self.G[s][a]
                Ts[s][a] = (Ss[s][a] + Ds[s][a]) / self.B[s][a]
                
        return Ts, Ps, As
    
    def _calcule_fixant_Ps_Ts_Ds(self, Ps, Ts, Ds):
        """
        Calcul à pension, cotisations et dépenses définis
        
        Paramètres :
            Ps : un dictionnaire, Ps[s][a] est le montant de la pension 
            par rapport aux salaires du scénario s à l'année a
            Ts : un dictionnaire, Ts[s][a] est le taux de cotisations 
            du scénario s à l'année a
            Ds : un dictionnaire, Ds[s][a] est le niveau de dépenses 
            du scénario s à l'année a
        """
    
        # Calcule As
        As = deepcopy(self.A)
        for s in self.scenarios:    
            for a in self.annees_futures:
                K = Ds[s][a] / self.B[s][a] / (Ps[s][a] + self.dP[s][a])
                As[s][a] = self.A[s][a] + (self.NR[s][a] - K * self.NC[s][a]) / (0.5*K + 1) / self.G[s][a]
                
        return Ts, Ps, As

    def _calcule_fixant_As_Ts_Ds(self, As, Ts, Ds):
        """
        Calcul à âge, cotisations et dépenses définis
        
        Paramètres :
            As : un dictionnaire, As[s][a] est l'âge de départ à la retraite 
            du scénario s à l'année a
            Ts : un dictionnaire, Ts[s][a] est le taux de cotisations 
            du scénario s à l'année a
            Ds : un dictionnaire, Ds[s][a] est le niveau de dépenses 
            du scénario s à l'année a
        """
    
        # Calcule Ps
        Ps = deepcopy(self.P)
        for s in self.scenarios:    
            for a in self.annees_futures:
                GdA = self.G[s][a] * ( As[s][a] - self.A[s][a] )
                K = (self.NR[s][a] - GdA) / (self.NC[s][a] + 0.5*GdA)
                Ps[s][a] = Ds[s][a] / self.B[s][a] / K - self.dP[s][a]
                
        return Ts, Ps, As
    
    def _calcule_fixant_As_RNV_S(self, As, RNVs, Ss):
        """
        Pilotage 1 : calcul à âge et niveau de vie défini
        
        Paramètres :
            As : un dictionnaire, As[s][a] est l'âge de départ à la retraite 
            du scénario s à l'année a
            RNVs : un dictionnaire, RNVs[s][a] est le niveau de vie 
            par rapport à l'ensemble de la population 
            du scénario s à l'année a
            Ss : un dictionnaire, Ss[s][a] est le solde financier 
            du scénario s à l'année a
        """
        
        # Calcule Ps et Ts
        Ts, Ps = deepcopy(self.T), deepcopy(self.P)
        for s in self.scenarios:    
            for a in self.annees_futures:                 
                GdA = self.G[s][a] * ( As[s][a] - self.A[s][a] )
                K = ( self.NR[s][a] - GdA ) / ( self.NC[s][a] + 0.5*GdA )
                Z = ( 1.0 - self.TCR[s][a] ) * self.CNV[s][a] / RNVs[s][a]
                U = 1.0 - ( self.TCS[s][a] - self.T[s][a] )
                L = Ss[s][a] / self.B[s][a]    
                Ps[s][a] = ( U - L - K*self.dP[s][a] ) / ( Z + K )
                Ts[s][a] = U - Ps[s][a] * Z 
                
        return Ts, Ps, As
    
        
    def _calcule_fixant_Ts_RNV_S(self, Ts, RNVs, Ss):
        """
        Pilotage 3 : calcul à cotisations et niveau de vie défini
        
        Paramètres :
            Ts : un dictionnaire, Ts[s][a] est le taux de cotisations
            du scénario s à l'année a
            RNVs : un dictionnaire, RNVs[s][a] est le niveau de vie 
            par rapport à l'ensemble de la population 
            du scénario s à l'année a
            Ss : un dictionnaire, Ss[s][a] est le solde financier 
            du scénario s à l'année a
        """
        
        # Calcule Ps et As
        Ps, As = deepcopy(self.P), deepcopy(self.A)    
        for s in self.scenarios:    
            for a in self.annees_futures:
                Ps[s][a] = RNVs[s][a] * (1-(self.TCS[s][a] + Ts[s][a]-self.T[s][a])) / self.CNV[s][a] / (1-self.TCR[s][a])
                K = (Ts[s][a] - Ss[s][a] / self.B[s][a]) / (Ps[s][a]+self.dP[s][a])
                As[s][a] = self.A[s][a] + ( self.NR[s][a] - K*self.NC[s][a] ) / (0.5*K + 1.0) / self.G[s][a]
                
        return Ts, Ps, As
    
    
    def _calcule_fixant_As_Ts_S(self, As, Ts, Ss):
        """
        Pilotage 4 : calcul à cotisations et âge définis
        
        Paramètres :
            As : un dictionnaire, As[s][a] est l'âge de départ à la retraite 
            du scénario s à l'année a
            Ts : un dictionnaire, Ts[s][a] est le taux de cotisations
            du scénario s à l'année a
            Ss : un dictionnaire, Ss[s][a] est le solde financier 
            du scénario s à l'année a
        """

        # Calcule Ps
        Ps = deepcopy(self.P)        
        for s in self.scenarios:    
            for a in self.annees_futures:
                GdA = self.G[s][a] * ( As[s][a]-self.A[s][a] )
                K = ( self.NR[s][a] - GdA ) / ( self.NC[s][a] + 0.5 * GdA )
                Ps[s][a] = (Ts[s][a]-Ss[s][a]/self.B[s][a])/K - self.dP[s][a]
                
        return Ts, Ps, As
        
    def _calcule_S_RNV_REV(self, Ts, Ps, As):
        """
        pilotage 0 : statu quo du COR
        Calcule les sorties du modèle de retraite en fonction des leviers.
        
        Paramètres :
            Ts : le taux de cotisations
            P : le niveau des pensions par rapport aux salaires
            A : âge moyen de départ à la retraite
        """
    
        S,RNV,REV, Depenses = dict(), dict(), dict(), dict()
    
        for s in self.scenarios:
    
            S[s], RNV[s], REV[s], Depenses[s] = dict(), dict(), dict(), dict()
    
            for a in self.annees:
    
                GdA = self.G[s][a] * ( As[s][a]-self.A[s][a] )
                K = ( self.NR[s][a] - GdA ) / ( self.NC[s][a] + 0.5*GdA )
                U = 1.0 - ( self.TCS[s][a] - self.T[s][a] )
                Depenses[s][a] = self.B[s][a] * K * ( Ps[s][a] + self.dP[s][a] )
                S[s][a] = self.B[s][a] * ( Ts[s][a] -  K * ( Ps[s][a] + self.dP[s][a] ) ) 
                RNV[s][a] =  Ps[s][a] * ( 1.0 - self.TCR[s][a] ) / (U - Ts[s][a]) * self.CNV[s][a]
    
                annee_naissance = round(a + 0.5 - As[s][a])
                age_mort = 60.0 + self.EV[s][annee_naissance]
                REV[s][a] = ( age_mort - As[s][a] ) / age_mort
    
        return S, RNV, REV, Depenses

    def genereTrajectoire(self, nom, valeur=None):
        """
        Crée une nouvelle trajectoire à partir de la valeur constante. 
        
        Paramètres :
            nom : une chaîne de caratères, le nom de la variable
            valeur : un flottant, la valeur numérique constante
        
        Description :        
            Retourne un dictionnaire contenant une trajectoire 
            dans tous les scénarios et pour toutes les années : 
            trajectoire[s][a] est la valeur numérique du scénario s à l'année a
            
            * Si la valeur n'est pas donnée, utilise par défaut la trajectoire du COR.
            * Si la valeur donnée est un flottant, utilise la trajectoire du 
            COR pour les années passées et cette valeur pour les années futures. 
            * Si la valeur donnée est un dictionnaire, considère que c'est 
            une trajectoire et utilise cette trajectoire.         
        
        Exemples        
            simulateur = SimulateurRetraites()
            simulateur.genereTrajectoire(simulateur, "A") # Départ à l'âge du COR
            simulateur.genereTrajectoire(simulateur, "A", 62.0) # Départ à 62.0 ans
            simulateur.genereTrajectoire(simulateur, "A", simulateur.A[1][2020]) # Départ à l'âge du COR en 2020
        """

        if type(valeur) is dict:
            # Si la valeur est un dictionnaire, on suppose que c'est une trajectoire 
            # et on le copie
            trajectoire = deepcopy(valeur)
        else:
            # Sinon, on suppose que c'est un flottant 
            # et on calcule la trajectoire du COR
            if (nom=="A"):
                trajectoire = deepcopy(self.A)
            elif (nom=="S"):
                S_COR, RNV_COR, REV_COR, Depenses_COR = self._calcule_S_RNV_REV(self.T, self.P, self.A)
                trajectoire = S_COR
            elif (nom=="P"):
                trajectoire = deepcopy(self.P)
            elif (nom=="T"):
                trajectoire = deepcopy(self.T)
            elif (nom=="RNV"):
                S_COR, RNV_COR, REV_COR, Depenses_COR = self._calcule_S_RNV_REV(self.T, self.P, self.A)
                trajectoire = RNV_COR
            elif (nom=="Depenses"):
                S_COR, RNV_COR, REV_COR, Depenses_COR = self._calcule_S_RNV_REV(self.T, self.P, self.A)
                trajectoire = Depenses_COR
            elif (nom=="REV"):
                S_COR, RNV_COR, REV_COR, Depenses_COR = self._calcule_S_RNV_REV(self.T, self.P, self.A)
                trajectoire = REV_COR
            else:
                raise TypeError('Mauvaise valeur pour le nom : %s' % (nom))
            
            if valeur is not None:
                # Propage la valeur constante dans la trajectoire
                # pour les années futures
                for s in self.scenarios:
                    for a in self.annees_futures:
                        trajectoire[s][a] = valeur

        return trajectoire

    def dessineConjoncture(self, taille_fonte_titre = 8, \
                  dessine_legende = False, scenarios_indices = None, 
                  dessine_annees = None):
        """
        Dessine les hypothèses de conjoncture.
        
        Paramètres :
            taille_fonte_titre : taille de la fonte du titre (par défaut, fs=8)
            dessine_legende : booleen, True si la légende doit être dessinée
            scenarios_indices : une liste d'entiers, la liste des indices des scénarios
            (par défaut, sc = range(1,7))
            dessine_annees : la liste des années à dessiner
        
        Exemple:
            simulateur = SimulateurRetraites()
            simulateur.dessineConjoncture()
        """
        pl.figure(figsize=(10,8))
        pl.suptitle(u"Projections du COR (hypothèses)",fontsize=16)
        for c in range(9):
            pl.subplot(3,3,c+1)
            nom = self.liste_variables[c]
            self.graphique(nom, taille_fonte_titre = taille_fonte_titre, \
                            dessine_legende = dessine_legende, \
                            scenarios_indices = scenarios_indices, \
                            dessine_annees = dessine_annees)
        pl.tight_layout(rect=[0, 0.03, 1, 0.95])
        return None
    
    def graphique(self, nom, v = None, taille_fonte_titre = 8, yaxis_lim = None, \
                  dessine_legende = False, scenarios_indices = None, 
                  dessine_annees = None):
        """
        Dessine un graphique associé à une variable donnée 
        pour tous les scénarios.
        
        Paramètres :
            nom : chaîne de caractère, nom de la variable
            v : variable à dessiner (par défaut, en fonction du nom)
            taille_fonte_titre : taille de la fonte du titre (par défaut, fs=8)
            yaxis_lim : une liste de taille 2, les bornes inférieures et supérieures 
            de l'axe des ordonnées (par défaut, utilise les paramètres de 
            l'objet)
            dessine_legende : booleen, True si la légende doit être dessinée
            scenarios_indices : une liste d'entiers, la liste des indices des scénarios
            (par défaut, sc = range(1,7))
            dessine_annees : la liste des années à dessiner
        
        Description :
            Le nom peut être une égal à une des chaînes de caractères parmi 
            les chaînes suivantes : "B", "NR", "NR", "G", "dP", "TPR", 
            "TPS", "CNV", "EV".
        
        Exemple :
            simulateur = SimulateurRetraites()
            simulateur.graphique("B")
        """
        if v is None:
            if nom=="B":
                v = self.B
            elif nom=="NR":
                v = self.NR
            elif nom=="NC":
                v = self.NC
            elif nom=="G":
                v = self.G
            elif nom=="dP":
                v = self.dP
            elif nom=="TPR":
                v = self.TCR
            elif nom=="TPS":
                v = self.TCS
            elif nom=="CNV":
                v = self.CNV
            elif nom=="EV":
                v = self.EV
            else:
                raise TypeError('Mauvaise valeur pour le nom : %s' % (nom))

        if scenarios_indices is None:
            scenarios_indices = self.scenarios
        
        if dessine_annees is not None:
            list_annees_dessin = dessine_annees
        else:
            if nom=="EV":
                list_annees_dessin=self.annees_EV
            else:
                list_annees_dessin=self.annees
    
        for s in scenarios_indices:
            if (nom=="RNV"):
                # Ce sont des % : multiplie par 100.0
                y = [ 100.0 * v[s][a] for a in list_annees_dessin ]
            else:
                y = [ v[s][a] for a in list_annees_dessin ]

            if (self.labels_is_long):
                label_variable = self.scenarios_labels[s]
            else:
                label_variable = self.scenarios_labels_courts[s]
            pl.plot(list_annees_dessin, y, label=label_variable )
    
        # titres des figures
        indice_variable = self.liste_variables.index(nom)
        titre_figure=self.liste_legendes[ indice_variable ]
           
        pl.title(titre_figure,fontsize=taille_fonte_titre)
        
        # Ajuste les limites de l'axe des ordonnées
        if yaxis_lim is None:
            # If the use did not set the yaxis_lim
            if nom in self.yaxis_lim.keys():
                # If the variable name was found in the dictionnary
                yaxis_lim = self.yaxis_lim[nom]
        else:
            pl.ylim(bottom=yaxis_lim[0],top=yaxis_lim[1])

        if dessine_legende:
            pl.legend(loc="best")
        return None
    
    def setAfficheMessageEcriture(self, affiche_quand_ecrit):
        """
        Configure l'affichage d'un message quand on écrit un fichier
        
        Paramètres :
            affiche_quand_ecrit : un booléen (par défaut = True)
        
        Exemple
            simulateur = SimulateurRetraites()
            simulateur.setAfficheMessageEcriture(False)
        """
        self.affiche_quand_ecrit = affiche_quand_ecrit
        return None
    
    def setImageFormats(self, ext_image):
        """
        Configure le format de sauvegarde des images
        
        Paramètres :
            ext_image : une liste de chaînes de caractères (par defaut, ext_image=["png","pdf"])
        
        Exemple :
            simulateur = SimulateurRetraites()
            simulateur.setImageFormats(["jpg"])
        """
        self.ext_image = ext_image
        return None

    def getImageFormats(self):
        """
        Retourne la liste des formats de sauvegarde des images.
        
        Exemple :
            simulateur = SimulateurRetraites()
            ext_image = simulateur.getImageFormats()
        """
        return self.ext_image 

    
    def setLabelLongs(self, labels_is_long):
        """
        Configure si la longueur des étiquettes est longue ou courte.
        
        Paramètres :
            labels_is_long : un booléen, True si les labels longs sont utilisés (par défaut = True)
        
        Exemple :
            simulateur = SimulateurRetraites()
            simulateur.setLabelLongs(False)
        """
        self.labels_is_long = labels_is_long
        return None

    def getLabelLongs(self):
        """
        Retourne le booléen associé à la longueur des étiquette. 
        
        Exemple :
            simulateur = SimulateurRetraites()
            labels_is_long = simulateur.getLabelLongs()
        """
        return self.labels_is_long 

    def setDirectoryImage(self, dir_image):
        """
        Configure le répertoire contenant les images
        
        Paramètres :
            dir_image : une chaîne de caractères, le répertoire contenant les images 
            (par défaut, dir_image="fig")
            exportées par sauveFigure.
        
        Exemple :
            simulateur = SimulateurRetraites()
            simulateur.setDirectoryImage("/tmp")
        """
        self.dir_image = dir_image
        return None

    def getDirectoryImage(self):
        """
        Retourne le répertoire contenant les images
        
        Exemple :
            simulateur = SimulateurRetraites()
            dir_image = simulateur.getDirectoryImage()
        """
        return self.dir_image

    def sauveFigure(self, f):
        """
        Sauvegarde l'image dans le répertoire
        
        Paramètres :
            f : une chaîne de caractères, le nom de base des fichiers à sauver
        
        Description :
            Sauvegarde l'image dans les formats définis. 
        
        Exemple :
            simulateur = SimulateurRetraites()
            simulateur.sauveFigure("conjoncture")
        """
    
        for ext in self.ext_image:
            basefilename = f + "." + ext
            filename = os.path.join(self.dir_image,basefilename)
            if self.affiche_quand_ecrit:
                print("Ecriture du fichier %s" % (filename))
            pl.savefig(filename)
        return None
    
    def dessineLegende(self):
        """
        Crée un graphique présentant les légendes des graphiques.
        
        Exemple :
            simulateur = SimulateurRetraites()
            simulateur.dessineLegende()
        """
        # Juste les légendes
        pl.figure(figsize=(6,2))
        for s in self.scenarios:
            pl.plot(0., 0., label = self.scenarios_labels[s])
        pl.legend(loc="center")
        pl.ylim(bottom=0.0, top=0.7)
        pl.axis('off')
        return None

    def _genereTrajectoirePIB(self):
        """
        Calcule le PIB dans les différents scénarios. 
        Source : https://fr.wikipedia.org/wiki/Produit_int%C3%A9rieur_brut_de_la_France
        """
        # Historique de PIBs (Milliards EUR)
        PIB_constate = {
        2005 : 1772.0, 
        2006 : 1853.3, 
        2007 : 1945.7, 
        2008 : 1995.8, 
        2009 : 1939.0, 
        2010 : 1998.5, 
        2011 : 2059.3, 
        2012 : 2091.1, 
        2013 : 2115.7, 
        2014 : 2141.1, 
        2015 : 2181.1, 
        2016 : 2228.9, 
        2017 : 2291.7, 
        2018 : 2353.1
        }
        # Croissance en fonction du scénario
        annee_dernier_PIB = 2018
        # Génère la trajectoire
        PIB = dict()
        for s in self.scenarios:
            PIB[s] = dict()
            croissance = self.scenarios_croissance[s]
            for a in self.annees:
                if (a<= annee_dernier_PIB):
                    PIB[s][a] = PIB_constate[a]
                else:
                    PIB[s][a] = (1.0 + croissance/100.0) * PIB[s][a - 1]
        return PIB

    def _calculePensionAnnuelleDroitDirect(self, PIB, As):
        """
        Calcule la pension annuelle de droit direct (brut) en kEUR.
        
        Paramètres :
            PIB : la trajectoire de PIB
            As : l'âge de départ à la retraite modifié par l'utilisateur
        """
        pensionBrut = dict()
        for s in self.scenarios:
            pensionBrut[s] = dict()
            for a in self.annees:
                GdA = self.G[s][a] * ( As[s][a] - self.A[s][a] )
                pensionBrut[s][a] = self.B[s][a] * self.P[s][a] * \
                    PIB[s][a] * 1000.0 / (self.NC[s][a] + 0.5 * GdA)
        return pensionBrut

    def calculeAge(self, REVcible):
        """
        Calcul de l'âge en fonction de la durée de vie à la retraite
        
        Paramètres :
            REVcible : la durée de vie à la retraite

        Description :
            Retourne un dictionnaire représentant une trajectoire d'âge de 
            départ en retraite.
            * Si la valeur cible n'est pas donnée, utilise par défaut la trajectoire du COR.
            * Si la valeur cible donnée est un flottant, utilise la trajectoire du 
            COR pour les années passées et cette valeur pour les années futures. 
            * Si la valeur cible donnée est un dictionnaire, considère que c'est 
            une trajectoire et utilise cette trajectoire. 
            
            Le calcul est réalisé par inversion numérique du ratio de durée  
            de vie en retraite. 
            Pour cela, nous utilisons le module scipy.optimize. 
            
            La trajectoire d'âge est uniquement déterminée par le ratio de durée 
            de vie en retraite. 
            C'est pourquoi on peut combiner la méthode calculeAge avec tout 
            pilotage prenant en entrée une trajectoire d'âge. 
            Par exemple, on peut combiner la méthode calculeAge avec la méthode 
            pilotageParPensionAgeCotisations. 

        Exemple :
            simulateur = SimulateurRetraites()
            REVcible = 0.30
            Acible = simulateur.calculeAge(REVcible = REVcible)
            analyse = simulateur.pilotageParSoldePensionAge(Acible = Acible)
        """

        def _EcartDeREV(As, args):
            """
            Pour un âge de départ à la retraite donné, 
            calcule la différence entre la durée de vie à la retraite pour 
            une année donnée et la durée de vie à la retraite 
            cible REVcible. 
            
            Paramètres :
                As : un flottant, l'âge de départ à la retraite
                args : une liste de quatre éléments [simulateur, annee, anneeReference, scenarioReference]
                simulateur : un SimulateurRetraite
                annee : un flottant, l'année du calcul de la durée de vie à la retraite
                anneeReference : un flottant, l'année de référence du calcul de la durée de vie à la retraite
                scenarioReference : un entier, le scénario de référence
                deltaREV : un flottant, la différence entre les deux durées de vie à la retraite
            
            Description :
                Utilise implicitement les variables simulateur et annee.
            """
            simulateur, scenario, annee, REVcible = args
            annee_naissance = round(annee + 0.5 - As)
            age_mort = 60.0 + simulateur.EV[scenario][annee_naissance]
            REV = ( age_mort - As ) / age_mort
            deltaREV = REVcible - REV
            return deltaREV

        REVs = self.genereTrajectoire("REV", REVcible)
        #
        As = deepcopy(self.A)
        for s in self.scenarios:
            for a in self.annees_futures:
                # Calcul l'âge
                args = [self, s, a, REVs[s][a]]
                result = spo.root_scalar(_EcartDeREV, \
                                         bracket= self.rechercheAgeBornes, \
                                         args = args, \
                                         rtol = self.rechercheAgeRTol)
                As[s][a] = result.root
        return As
